#include <stdlib.h>
#include <time.h>
#include <conio.h>
#include "cga.h"

enum {EMPTY, MONSTER, WALL, GOLD, PLAYER, UNKNOWN};

extern unsigned char THETILES[];

char map[10][20], hidden[10][20], valid[10][20];
int px, py, replay = 1, playing, gold = 0;

#define MIN(a,b)  ((a)<(b)?(a):(b))
#define MAX(a,b)  ((a)>(b)?(a):(b))

void show_ending();
void draw_tile(int x, int y);

int chancer(void)
{
    int cnt = 0, i;
    for (i=0; i < 100; i++)
        if (rand() > RAND_MAX/2) cnt++;
    return cnt;
}

void fill_valid(int x, int y)
{
    if (x < 0 || y < 0 || x > 19 || y > 9) return;
    if (!valid[y][x] && map[y][x] == EMPTY) {
        valid[y][x] = 1;
        fill_valid(x + 1, y);
        fill_valid(x - 1, y);
        fill_valid(x, y + 1);
        fill_valid(x, y - 1);
        fill_valid(x + 1, y - 1);
        fill_valid(x - 1, y - 1);
        fill_valid(x + 1, y + 1);
        fill_valid(x - 1, y + 1);
    }
}

int valid_count(void)
{
    int x, y, count = 0;
    for (y=0; y < 10; y++)
        for (x=0; x < 20; x++)
            if (valid[y][x]) count++;
    return count;
}

void place_random(int c, int cf)
{
    while (1) {
        int x = rand()%20;
        int y = rand()%10;
        if ((cf || valid[y][x]) && map[y][x] == EMPTY) {
            map[y][x] = c;
            valid[y][x] = 0;
            return;
        }
    }
}

void place_again(int c)
{
    int needhidden = 0, x, y;
    for (y=0; y<10; y++)
        for (x=0; x<10; x++)
            if (valid[y][x] && hidden[y][x] && map[y][x] == EMPTY) {
                needhidden = 1;
                break;
            }
    while (1) {
        int x = rand()%20;
        int y = rand()%10;
        if (valid[y][x] && map[y][x] == EMPTY && hidden[y][x] == needhidden) {
            map[y][x] = c;
            draw_tile(x, y);
            return;
        }
    }
}

void map_init(void)
{
    int x, y, valids = 0;
    cga_msg(1, "Generating map...");
    while (valids < 100) {
        memset(map, 0, sizeof(map));
        memset(valid, 0, sizeof(valid));
        memset(hidden, 1, sizeof(hidden));
        for (x=0; x < 70; x++) place_random(WALL, 1);
        py = 3 + rand()%6;
        px = 5 + rand()%10;
        map[py][px] = EMPTY;
        fill_valid(px, py);
        map[py][px] = PLAYER;
        valid[py][px] = 0;
        valids = valid_count();
    }
    for (x=0; x < 30; x++) place_random(MONSTER, 0);
    for (x=0; x < 13; x++) place_random(GOLD, 0);
    hidden[py][px] = 0;
}

void draw_tile(int x, int y)
{
    if (hidden[y][x]) cga_tile(x*16, y*16, TILEPTR(THETILES, UNKNOWN));
    else cga_tile(x*16, y*16, TILEPTR(THETILES, map[y][x]));
}

void draw_board(void)
{
    int x, y;
    for (y=0; y < 10; y++)
        for (x=0; x < 20; x++)
            draw_tile(x, y);
}

void draw_bar(int y)
{
    int x;
    for (x=1; x < 19; x++) cga_tile(x*16, y, TILEPTR(THETILES, 7));
    cga_tile(0, y, TILEPTR(THETILES, 6));
    cga_tile(304, y, TILEPTR(THETILES, 8));
}

void draw_status(void)
{
    static char status[80];
    sprintf(status, "                   Found %02i of 13 gold", gold);
    cga_msg(23, status);
}

void draw_all(void)
{
    draw_board();
    draw_bar(160);
    draw_status();
}

void reveal(int x, int y)
{
    hidden[y][x] = 0;
    draw_tile(x, y);
}

void reveal_show(int x, int y)
{
    int real = map[y][x];
    reveal(x, y);
    delay(200);
    map[y][x] = EMPTY;
    draw_tile(x, y);
    delay(200);
    map[y][x] = real;
    draw_tile(x, y);
    delay(200);
    map[y][x] = EMPTY;
    draw_tile(x, y);
    delay(200);
    map[y][x] = real;
    draw_tile(x, y);
    delay(200);
    map[y][x] = EMPTY;
    draw_tile(x, y);
    delay(200);
    map[y][x] = real;
    draw_tile(x, y);
    delay(200);
}

void move_to(int nx, int ny)
{
        hidden[ny][nx] = 0;
    map[ny][nx] = PLAYER;
    map[py][px] = EMPTY;
    draw_tile(nx, ny);
    draw_tile(px, py);
    px = nx;
    py = ny;
}

int getkey(void)
{
    int key = getch();
    if (key == 0) key = 1000 + getch();
    return key;
}

void killkey(void)
{
    while (kbhit()) getkey();
}

void msgline(const char* msg)
{
    cga_msg(25, msg);
    killkey();
    getkey();
    cga_msg(25, "");
}

int ask(const char* msg)
{
    killkey();
    while (1) {
        int key;
        cga_msg(25, msg);
        key = getkey();
        cga_msg(25, "");
        if (key == 'y' || key == 'Y') return 1;
        if (key == 'n' || key == 'N' || key == 27) return 0;
        cga_msg(25, "Invalid key, please press Y or N");
        delay(300);
    }
}

void encounter(void)
{
    int key, chance, loop = 1;
    killkey();
    while (loop) {
        char tmp[40];
        cga_msg(23, "A monster is blocking your path!");
        chance = 100 - (15+6*gold);
        sprintf(tmp, "(R)un (%02i%% chance) (D)rop gold (Q)uit", chance);
        cga_msg(24, tmp);
        key = getkey();
        cga_msg(23, "");
        cga_msg(24, "");
        switch (key) {
        case 'R':
        case 'r':
            cga_msg(24, "You attempt to run away");delay(300);
            cga_msg(24, "You attempt to run away.");delay(300);
            cga_msg(24, "You attempt to run away..");delay(300);
            cga_msg(24, "You attempt to run away...");delay(300);
            cga_msg(24, "");delay(400);
            if (chancer() < chance) {
                cga_msg(24, "You managed to run away!");
                getkey();
                loop = 0;
            } else {
                cga_msg(24, "Your attempt to run away failed...");
                getkey();
                loop = playing = 0;
                cga_msg(23, "  The gold you carried was enough to");
                cga_msg(24, "slow you down and provide a nice meal");
                replay = ask("for the cave monster.  Retry? (Y/N)");
            }
            break;
        case 'D':
        case 'd':
            if (gold > 0) {
                gold--;
                place_again(GOLD);
            }
            else msgline("You have no gold to drop");
            break;
        case 'Q':
        case 'q':
            loop = playing = 0;
            break;
        default:
            msgline("Invalid key...");
            break;
        }
        killkey();
    }
}

void move_towards(int dx, int dy)
{
    if (px + dx < 0 || py + dy < 0 || px + dx > 19 || py + dy > 9) {
        msgline("Cannot move outside the map...");
        return;
    }
    switch (map[py + dy][px + dx]) {
    case WALL:
        reveal(px + dx, py + dy);
        msgline("There is a wall there...");
        return;
    case GOLD:
        gold++;
        draw_status();
        cga_msg(24, "You found gold!");
        reveal_show(px + dx, py + dy);
        if (gold == 13) {
            show_ending();
            playing = 0;
        }
        break;
    case MONSTER:
        cga_msg(24, "Oh no!");
        reveal_show(px + dx, py + dy);
        encounter();
        if (playing) {
            map[py + dy][px + dx] = EMPTY;
            reveal(px + dx, py + dy);
        }
        draw_status();
        return;
    }
    move_to(px + dx, py + dy);
}

void scanfor(int c, const char* ok, const char* fail)
{
    int found = 0, x1 = MAX(px - 1, 0), y1 = MAX(py - 1, 0), x2 = MIN(19, px + 1), y2 = MAX(9, py + 1);
    int x, y;
    for (y=y1; y<=y2; y++)
        for (x=x1; x<=x2; x++)
            if (map[y][x] == c) {
                found = 1;
                break;
            }
    msgline(found ? ok : fail);
}

void play(void)
{
    int key;
    draw_all();
    while (playing) {
        cga_msg(24, "(1-8) Move (L)isten (E)xamine (Q)uit");
        key = getkey();
        cga_msg(24, "");
        switch (key) {
        case 'E':
        case 'e':
            scanfor(GOLD, "Something is sparkling...", "There is nothing here...");
            break;
        case 'L':
        case 'l':
            scanfor(MONSTER, "You can hear some growling sounds...", "You cannot hear anything...");
            break;
        case 27:
        case 'Q':
        case 'q':
            playing = !ask("Do you want to quit (Y/N)?");
            break;
        case '1':
            move_towards(-1, 1);
            break;
        case '2':
            move_towards(0, 1);
            break;
        case '3':
            move_towards(1, 1);
            break;
        case '4':
            move_towards(-1, 0);
            break;
        case '6':
            move_towards(1, 0);
            break;
        case '7':
            move_towards(-1, -1);
            break;
        case '8':
            move_towards(0, -1);
            break;
        case '9':
            move_towards(1, -1);
            break;
        default:
            msgline("Invalid key...");
        }
        killkey();
    }
}

void show_intro(void)
{
    memset(cgadata, 0, 320*200/4);
    draw_bar(0);
    draw_bar(176);
    cga_msg( 4, " *** COLOSSAL GRAND ADVENTURE CAVE ***");
    cga_msg( 6, "   Welcome to Colossal Grand Adventure");
    cga_msg( 7, " Cave (or simply CGA Cave).  Your goal");
    cga_msg( 8, " in this game is to find all 13 pieces");
    cga_msg( 9, " of gold hidden in the cave and avoid");
    cga_msg(10, " the monsters looming in the darkness.");
    cga_msg(12, " Use the following keys to navigate:");
    cga_msg(15, "    7   8   9   [E] Examine the area");
    cga_msg(16, "     \\  |  /        around the player");
    cga_msg(17, "      \\ | /");
    cga_msg(18, "    4-- * --6   [L] Listen for noises");
    cga_msg(19, "      / | \\         from monsters");
    cga_msg(20, "     /  |  \\");
    cga_msg(21, "    1   2   3   [Q] Quit the game");
    cga_msg(25, "Press a key, any key, to continue...");
    getkey();
    memset(cgadata, 0, 320*200/4);
}

void show_ending(void)
{
    memset(cgadata, 0, 320*200/4);
    draw_bar(0);
    draw_bar(176);
    cga_msg( 7, "      * * * CONGRATULATIONS * * *");
    cga_msg(12, "   You managed to explore the cave (or");
    cga_msg(13, " at least the parts that matter) while");
    cga_msg(14, " avoiding to become monster snack.");
    cga_msg(16, "   Your efforts will be rewarded with");
    cga_msg(17, " all the gold you have collected.  How");
    cga_msg(18, " does it feel to be a rich person?");
    replay = ask("Do you want to play again?");
    memset(cgadata, 0, 320*200/4);
}

int main()
{
    cga_init();
    while (replay) {
        replay = gold = 0;
        playing = 1;
        srand(time(0));
        show_intro();
        map_init();
        play();
    }
    cga_shutdown();
    return 0;
}
